/* sel-queries.c
 *
 * Copyright (C) 2004 Vivien Malerba
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 */

#include "sel-queries.h"
#include <libgnomedb/libgnomedb.h>
#include "gnome-db-query.h"
#include "gnome-db-qf-all.h"
#include "gnome-db-qf-field.h"
#include "gnome-db-qf-value.h"
#include "sel-onequery.h"

/*
 *
 * Module for all queries managed by GnomeDbDict
 *
 */
static void         module_queries_fill_model (Module *module);
static void         module_queries_free (Module *module);
static const gchar *module_queries_col_name (Module *module, guint colno);
static Module      *module_queries_obj_manager (Module *module, GtkTreeIter *iter, GObject *object);
static void         module_queries_model_store_data (Module *module, GtkTreeIter *iter);

Module *
sel_module_queries_new (GnomeDbSelector *mgsel, gboolean insert_header, 
			GtkTreeIter *iter, gpointer data)
{
	Module *module;

	module = g_new0 (Module, 1);
	module->selector = mgsel;
	module->fill_model = module_queries_fill_model;
	module->free = module_queries_free;
	module->col_name = module_queries_col_name;
	module->obj_manager = module_queries_obj_manager;
	module->model_store_data = module_queries_model_store_data;
	module->mod_data = NULL;
	module->iter = NULL;
	module->parent_module = NULL;
	module->sub_modules = NULL;

	if (insert_header) {
		GdkPixbuf *pixbuf = NULL;
		GtkTreeModel *model = mgsel->priv->model;

		pixbuf = gnome_db_stock_get_icon_pixbuf (GNOME_DB_STOCK_QUERY);
		module->iter = g_new0 (GtkTreeIter, 1);
		gtk_tree_store_append (GTK_TREE_STORE (model), module->iter, iter);
		gtk_tree_store_set (GTK_TREE_STORE (model), module->iter, NAME_COLUMN, 
				    _("Queries"), PIXBUF_COLUMN, pixbuf, 
				    CONTENTS_COLUMN, CONTENTS_TOP_CATEGORY, 
				    SUB_MODULE_COLUMN, NULL, -1);		
	}
	else {
		if (iter)
			module->iter = gtk_tree_iter_copy (iter);
	}

	return module;
}


static GSList *module_queries_get_objects_list (Module *module);
static void
module_queries_fill_model (Module *module)
{
	GtkTreeModel *model;
	GdkPixbuf *pixbuf_query = NULL;

	pixbuf_query = gnome_db_stock_get_icon_pixbuf_file ("gnome-db-query_16x16.png");

	/* Module's private data */
	module->mod_data = g_new0 (ModFlatData, 1);
	FLAT_DATA (module)->manager = NULL;
	FLAT_DATA (module)->manager_weak_refed = FALSE;
	FLAT_DATA (module)->fallback_obj_pixbuf = pixbuf_query;
	FLAT_DATA (module)->get_objects_list = module_queries_get_objects_list;

	/* Initial model filling */
	model = module->selector->priv->model;
	flat_init_model_fill (module, model);

	/* Signals handlers */
	g_signal_connect (G_OBJECT (module->selector->priv->dict), "query_added",
			  G_CALLBACK (flat_obj_added_cb), module);
	g_signal_connect (G_OBJECT (module->selector->priv->dict), "query_removed",
			  G_CALLBACK (flat_obj_removed_cb), module);
	g_signal_connect (G_OBJECT (module->selector->priv->dict), "query_updated",
			  G_CALLBACK (flat_obj_updated_cb), module);
}

static GSList *
module_queries_get_objects_list (Module *module)
{
	return gnome_db_dict_get_queries (module->selector->priv->dict);
}

static void
module_queries_free (Module *module)
{
	GSList *list = module->sub_modules;

	/* Free the sub modules */
	while (list) {
		(MODULE (list->data)->free) (MODULE (list->data));
		g_free (list->data);
		list = g_slist_next (list);
	}
	if (module->sub_modules) {
		g_slist_free (module->sub_modules);
		module->sub_modules = NULL;
	}

	/* free this module */
	g_signal_handlers_disconnect_by_func (G_OBJECT (module->selector->priv->dict),
					      G_CALLBACK (flat_obj_added_cb), module);
	g_signal_handlers_disconnect_by_func (G_OBJECT (module->selector->priv->dict),
					      G_CALLBACK (flat_obj_removed_cb), module);
	g_signal_handlers_disconnect_by_func (G_OBJECT (module->selector->priv->dict),
					      G_CALLBACK (flat_obj_updated_cb), module);

	if (module->iter)
		gtk_tree_iter_free (module->iter);

	flat_free_mod_data (module);
	g_free (module->mod_data);
	module->mod_data = NULL;
}


static const gchar *
module_queries_col_name (Module *module, guint colno)
{
	switch (colno) {
	case 0:
		return _("Query");
		break;
	case EXTRA1_COLUMN:
		return _("Type");
		break;
	default:
		return NULL;
		break;
	}
}

/*
 * module_queries_obj_manager
 *
 * This function is called when a new object is created (GnomeDbQuery). It simply creates a new
 * Module (which will be appended to the list of modules of the module for the queries) for 
 * for the contents of that new query.
 */
static Module *
module_queries_obj_manager (Module *module, GtkTreeIter *iter, GObject *object)
{
	Module *sub_module = NULL;

	g_assert (object && IS_GNOME_DB_QUERY (object));

	if ((module->selector->priv->mode & GNOME_DB_SELECTOR_QVIS_FIELDS) ||
	    (module->selector->priv->mode & GNOME_DB_SELECTOR_QALL_FIELDS) ||
	    (module->selector->priv->mode & GNOME_DB_SELECTOR_TARGETS) ||
	    (module->selector->priv->mode & GNOME_DB_SELECTOR_JOINS) ||
	    (module->selector->priv->mode & GNOME_DB_SELECTOR_SUB_QUERIES))
		sub_module = sel_module_onequery_new (module->selector, FALSE, iter, object);

	return sub_module;
}

static void
module_queries_model_store_data (Module *module, GtkTreeIter *iter)
{
	GObject *obj;
	GtkTreeModel *model;

	model = module->selector->priv->model;
	gtk_tree_model_get (model, iter, OBJ_COLUMN, &obj, -1);

	if (obj && IS_GNOME_DB_QUERY (obj)) {
		const gchar *str1;
		gtk_tree_store_set (GTK_TREE_STORE (model), iter, 
				    EXTRA1_COLUMN, gnome_db_query_get_query_type_string (GNOME_DB_QUERY (obj)),
				    -1);
		str1 = gnome_db_base_get_name (GNOME_DB_BASE (obj));
		if (!str1 || !(*str1)) {
			gtk_tree_store_set (GTK_TREE_STORE (model), iter, 
				    NAME_COLUMN, _("Query <no name>"),
				    -1);
		}
	}
}
